import math
import operator
from datetime import date, datetime, timedelta
from functools import reduce
from typing import Generator, List, Optional, Tuple, Type, TypeVar

import pandas as pd
from celery import shared_task
from colorfield.fields import ColorField
from django.contrib import admin
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group, Permission
from django.contrib.contenttypes.models import ContentType
from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import models
from django.db.models import (
    Case,
    Count,
    Exists,
    ExpressionWrapper,
    F,
    OuterRef,
    Q,
    QuerySet,
    Subquery,
    Sum,
    Value,
    When,
)
from django.db.models.fields import FloatField
from django.db.models.functions import Ceil, Coalesce
from django.db.models.signals import m2m_changed, post_delete, post_save
from django.db.utils import ProgrammingError
from django.dispatch import receiver
from django.utils.timezone import make_naive
from django.utils.translation import gettext
from django.utils.translation import gettext_lazy as _
from dynamic_preferences.registries import global_preferences_registry
from mptt.models import MPTTModel, TreeForeignKey
from psycopg.types.range import TimestamptzRange
from wbcore.contrib.directory.models import (
    Company,
    EmployerEmployeeRelationship,
    Person,
)
from wbcore.contrib.directory.models import Position as CRMPosition
from wbcore.contrib.directory.signals import deactivate_profile
from wbcore.contrib.notifications.dispatch import send_notification
from wbcore.contrib.notifications.utils import create_notification_type
from wbcore.models import WBModel
from wbcore.models.fields import YearField
from wbcore.utils.models import ComplexToStringMixin

from .absence import AbsenceRequest, AbsenceRequestPeriods
from .calendars import DayOff, DayOffCalendar, DefaultDailyPeriod
from .preferences import (
    default_vacation_days_per_year,
    get_is_external_considered_as_internal,
    get_main_company,
    get_previous_year_balance_expiration_date,
    long_vacation_number_of_days,
)

User = get_user_model()


class ActiveEmployeeManager(models.Manager):
    """Custom Manager for filtering directly Active Employees. Exclude objects without reverse related field user_account and profile"""

    def __init__(self, only_internal=False, **kwargs):
        self.only_internal = only_internal
        super().__init__(**kwargs)

    def get_queryset(self) -> "models.QuerySet[EmployeeHumanResource]":
        contract_type_condition = (
            Q(contract_type=EmployeeHumanResource.ContractType.INTERNAL)
            if (self.only_internal and not get_is_external_considered_as_internal())
            else Q(contract_type__isnull=False)
        )
        return (
            super()
            .get_queryset()
            .filter(
                Q(is_active=True)
                & Q(profile__isnull=False)
                & Q(profile__user_account__isnull=False)
                & contract_type_condition
            )
        )


SelfEmployeeHumanResource = TypeVar("SelfEmployeeHumanResource", bound="EmployeeHumanResource")


class EmployeeHumanResource(ComplexToStringMixin, WBModel):
    """
    Stores a single Employee entry, related to :model:`directory.Person`.
    """

    class ContractType(models.TextChoices):
        INTERNAL = "INTERNAL", _("Internal")
        EXTERNAL = "EXTERNAL", _("External")

    class ExtraDaysBalanceFrequency(models.TextChoices):
        MONTHLY = "MONTHLY", _("Monthly")
        YEARLY = "YEARLY", _("Yearly")

        def get_pandas_frequency(self) -> str:
            """
            Return the pandas frequency representation of this frequency
            """
            if self.value == self.MONTHLY:
                return "M"
            if self.value == self.YEARLY:
                return "Y"

        def get_date_range(self, _d: datetime) -> tuple[datetime, datetime]:
            """
            Return a tuple of datetime range representing this frequency
            """
            if self.value == self.MONTHLY:
                return ((_d - pd.tseries.offsets.MonthBegin(1)).to_pydatetime(), _d)
            if self.value == self.YEARLY:
                return (
                    (_d - pd.tseries.offsets.YearEnd(1)).to_pydatetime() + timedelta(days=1),
                    (_d + pd.tseries.offsets.YearBegin(1)).to_pydatetime() - timedelta(days=1),
                )

        def get_yearly_periods_count(self) -> int:
            """
            Return the number of yearly periods defined by this frequency
            """
            if self.value == self.MONTHLY:
                return 12
            if self.value == self.YEARLY:
                return 1

        def get_period_index(self, _d: datetime) -> int:
            """
            Get the period index
            """
            if self.value == self.MONTHLY:
                return _d.month
            if self.value == self.YEARLY:
                return 1

    profile = models.OneToOneField(
        "directory.Person",
        related_name="human_resources",
        on_delete=models.CASCADE,
        verbose_name=_("Employee"),
        help_text=_("The CRM profile related to this employee"),
    )
    is_active = models.BooleanField(
        verbose_name=_("Is active"),
        help_text=_(
            "If false, the employee will be considered as not active but his absence requests will be preserved"
        ),
        default=True,
    )
    extra_days_frequency = models.CharField(
        max_length=16,
        default=ExtraDaysBalanceFrequency.YEARLY,
        choices=ExtraDaysBalanceFrequency.choices,
        verbose_name=_("Extra Days Frequency"),
        help_text=_(
            "The frequency at which an additional number of vacation days is enabled for this employee (defaults to yearly)"
        ),
    )

    occupancy_rate = models.FloatField(
        verbose_name=_("Occupation Rate"),
        help_text=_("The occupation rate in percent, 100% being employed fulltime"),
        default=1,
    )

    contract_type = models.CharField(
        max_length=16,
        default=ContractType.INTERNAL,
        choices=ContractType.choices,
        verbose_name=_("Employee Type"),
        help_text=_(
            "If Internal, the employee is considered a full-time employee and thus has employee access to the system."
        ),
    )
    position = models.ForeignKey(
        "wbhuman_resources.Position",
        related_name="employees",
        on_delete=models.SET_NULL,
        verbose_name=_("Position"),
        limit_choices_to=models.Q(height=0),
        null=True,
        blank=True,
        help_text=_("The position this employee belongs to"),
    )
    direct_manager = models.ForeignKey(
        "directory.Person",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        verbose_name=_("Direct Manager"),
        related_name="managed_employees",
    )
    calendar = models.ForeignKey(
        to="wbhuman_resources.DayOffCalendar",
        related_name="employees",
        on_delete=models.PROTECT,
    )
    weekly_off_periods = models.ManyToManyField(
        to="wbhuman_resources.DefaultDailyPeriod",
        through="wbhuman_resources.EmployeeWeeklyOffPeriods",
        through_fields=("employee", "period"),
        related_name="employees_off",
        verbose_name=_("Weekly off periods"),
    )
    objects = models.Manager()
    active_internal_employees = ActiveEmployeeManager(only_internal=True)
    active_employees = ActiveEmployeeManager()

    enrollment_at = models.DateField(verbose_name=_("Enrolled at"))
    disenrollment_at = models.DateField(verbose_name="Disenroll at", blank=True, null=True)

    class Meta:
        verbose_name = _("Employee Human Resource")
        verbose_name_plural = _("Employee Human Resources")
        permissions = [("administrate_employeehumanresource", "Can administrate Employee Human Resource")]
        notification_types = [
            create_notification_type(
                code="wbhuman_resources.employeehumanresource.deactivate",
                title="Deactivate Employee Notification",
                help_text="Notify the requester when an employee has been successfully deactivated",
            ),
            create_notification_type(
                code="wbhuman_resources.employeehumanresource.vacation",
                title="Vacation Notification",
                help_text="Notifies you when there are Vacation days that you still have to take",
            ),
            create_notification_type(
                code="wbhuman_resources.employeehumanresource.birthday",
                title="Birthday Notification",
                help_text="Notifies you when today one of your colleagues is celebrating their birthday",
            ),
        ]

    def unassign_position_groups(self):
        """
        Un-assign the employee to the belonged human resource groups
        """
        if (user := getattr(self.profile, "user_account", None)) and self.position:
            for position in self.position.get_ancestors(include_self=True):
                for group in position.groups.all():
                    user.groups.remove(group)

    def assign_position_groups(self):
        """
        Assign the employee to the position permission group
        """
        if (user := getattr(self.profile, "user_account", None)) and self.position:
            for position in self.position.get_ancestors(include_self=True):
                for group in position.groups.difference(user.groups.all()):
                    user.groups.add(group)

    def deactivate(
        self, substitute: Optional[models.Model] = None, disenrollment_date: date | None = None
    ) -> List[str]:
        """
        Utility method to deactivate/disenroll an employee

        Trigger a signal "deactivate_profile" that every module can implement in order to define module level business logic for employee disenrollment.

        Args:
            substitute:
            disenrollment_date:

        Returns:
            A list of message feedback received from the potential receivers
        """
        if not disenrollment_date:
            disenrollment_date = date.today()
        # send signal to deactivate employee
        self.disenrollment_at = disenrollment_date
        res = deactivate_profile.send(self.__class__, instance=self.profile, substitute_profile=substitute)

        self.is_active = False
        self.save()
        self.profile.user_account.is_active = False
        self.profile.user_account.save()
        self.assign_vacation_allowance_from_range(
            self.enrollment_at, self.disenrollment_at
        )  # recompute and close possible open balance
        try:
            main_company = Company.objects.get(id=get_main_company())
            self.profile.employers.remove(main_company)
        except Company.DoesNotExist:
            pass
        return [msg for _, msg in res]

    def get_managed_employees(self, include_self: bool | None = True) -> "QuerySet[SelfEmployeeHumanResource]":
        """
        Returns all the direct managed employees, from the current position and all its descendants
        Args:
            include_self: Set to False if the returned queryset needs to exclude the employee. Default to False

        Returns:
            A queryset of EmployeeHumanResources
        """
        if (user := self.profile.user_account) and user.has_perm("wbhuman_resources.administrate_absencerequest"):
            qs = EmployeeHumanResource.active_employees.all()
        else:
            conditions = [Q(direct_manager=self.profile), Q(id=self.id)] + [
                Q(position__in=position.get_descendants(include_self=True))
                for position in self.profile.managed_positions.all()
            ]
            qs = EmployeeHumanResource.active_employees.filter(reduce(operator.or_, conditions))
        if not include_self:
            qs = qs.exclude(id=self.id)
        return qs.distinct()

    def get_managers(self, only_direct_manager: bool = False) -> Generator[SelfEmployeeHumanResource, None, None]:
        """
        Returns the direct manager of this employee and the potential global manager. Prioritize the direct manager, and if not available, the next position manager in the company hierarchy

        Returns:
            A generator yielding managers
        """
        if direct_manager := self.direct_manager:
            yield direct_manager
        elif main_position := self.position:
            if (
                next_position_with_manager := main_position.get_ancestors(ascending=True, include_self=True)
                .filter(manager__isnull=False)
                .first()
            ):
                yield next_position_with_manager.manager
        if not only_direct_manager:
            global_manager_permission = Permission.objects.get(
                codename="administrate_employeehumanresource", content_type=ContentType.objects.get_for_model(self)
            )
            for global_manager_user in global_manager_permission.user_set.filter(profile__isnull=False):
                yield global_manager_user.profile

    def is_manager_of(self, administree: SelfEmployeeHumanResource, include_self: bool = False) -> bool:
        """
        Return true if self is the manager of administree
        Args:
            administree: The employee to check hierarchy against

        Returns:
            A boolean
        """
        return self.get_managed_employees(include_self=include_self).filter(id=administree.id).exists()

    def compute_str(self) -> str:
        return f"{self.profile.first_name} {self.profile.last_name}"

    def __str__(self) -> str:
        return self.computed_str

    def save(self, *args, **kwargs):
        if not self.calendar:
            self.calendar = global_preferences_registry.manager()["wbhuman_resources__employee_default_calendar"]
        if not self.enrollment_at:
            self.enrollment_at = date.today()
        super().save(*args, **kwargs)

    def extract_workable_periods(
        self, start: datetime, end: datetime, count: int = None
    ) -> Generator[Tuple[datetime, datetime, DefaultDailyPeriod], None, None]:
        """
        Utility function that returns the day off for an employee, including the calendar day off and the possible weekly day off
        Args:
            start_date: lower bound range
            end_date:  upper bound range
            count: if specified, will stop the iteration at the specified index
        Returns:
            a list of datetime range representing each a period where the employee if off
        """
        cursor = start.date()
        index = 0
        while cursor <= end.date() and (not count or index < count):
            for period in self.calendar.default_periods.order_by("lower_time"):
                lower_datetime = period.get_lower_datetime(cursor)
                upper_datetime = period.get_upper_datetime(cursor)
                if (
                    lower_datetime >= start
                    and upper_datetime <= end
                    and not EmployeeWeeklyOffPeriods.objects.filter(
                        weekday=cursor.weekday(), period=period, employee=self
                    ).exists()
                    and not self.calendar.days_off.filter(date=cursor, count_as_holiday=True).exists()
                ):
                    yield cursor, period
                    index += 1
            cursor += timedelta(days=1)

    def assign_vacation_allowance_from_range(self, from_date: date, to_date: date):
        """
        Assign the proper monthly allowance from the given range if it doesn't yet exist.

        Assign partial allowance if the date range don't span a full period. Check also if the given range don't over span the enrollment and disenrollment employee's date

        Args:
            from_date: start date range
            to_date: end date range
        """
        from_date = max(self.enrollment_at, from_date)
        if self.disenrollment_at:
            to_date = min(self.disenrollment_at, to_date)
        periods = self.calendar.default_periods.order_by("lower_time")
        if periods.exists():
            # Convert to datetime
            from_datetime = periods.first().get_lower_datetime(from_date)
            to_datetime = periods.last().get_upper_datetime(to_date)

            frequency = self.ExtraDaysBalanceFrequency[self.extra_days_frequency]

            period_base_allowance = (
                default_vacation_days_per_year()
                * self.calendar.get_daily_hours()
                / frequency.get_yearly_periods_count()
            )
            for _d in pd.date_range(from_datetime, to_datetime, freq=frequency.get_pandas_frequency()):
                current_year_balance = self.get_or_create_balance(_d.year)[0]
                [start_period, end_period] = frequency.get_date_range(_d)
                total_workable_periods_count = len(list(self.extract_workable_periods(start_period, end_period)))
                actual_workable_periods_count = len(
                    list(self.extract_workable_periods(max(start_period, from_datetime), min(end_period, to_datetime)))
                )
                if actual_workable_periods_count and total_workable_periods_count:
                    BalanceHourlyAllowance.objects.get_or_create(
                        balance=current_year_balance,
                        period_index=frequency.get_period_index(_d),
                        defaults={
                            "hourly_allowance": period_base_allowance
                            * actual_workable_periods_count
                            / total_workable_periods_count
                        },
                    )

    def get_or_create_balance(self, year: int) -> tuple["EmployeeYearBalance", bool]:
        """
        Wrapper around get_or_create

        Args:
            year: lookup year argument

        Returns:
            A tuple of EmployeeYearBalance and bool. Boolean is true if the balance was actually created.
        """
        return EmployeeYearBalance.objects.get_or_create(employee=self, year=year)

    @classmethod
    def get_endpoint_basename(cls) -> str:
        return "wbhuman_resources:employee"

    @classmethod
    def get_representation_endpoint(cls) -> str:
        return "wbhuman_resources:employeehumanresourcerepresentation-list"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"

    @classmethod
    def get_representation_label_key(cls) -> str:
        return "{{ computed_str }}"

    @classmethod
    def annotated_queryset(
        cls, qs: QuerySet[SelfEmployeeHumanResource], end_of_month: date
    ) -> QuerySet[SelfEmployeeHumanResource]:
        """
        Utility classmethod to annotate the employee queryset with a set of usage and balance statistics variables

        Args:
            qs: The queryset to annotate
            end_of_month: Date at which the absence periods will be exclude from the balance usage

        Returns:
            The annotated queryset
        """
        long_vacation_days = long_vacation_number_of_days()
        vacation_days_per_year = default_vacation_days_per_year()
        return qs.annotate(
            x1=Case(
                When(
                    extra_days_frequency=EmployeeHumanResource.ExtraDaysBalanceFrequency.YEARLY.name,
                    then=F("occupancy_rate") * vacation_days_per_year * 2,
                ),
                default=F("occupancy_rate") * vacation_days_per_year / 6,
                output_field=FloatField(),
            ),
            extra_days_per_period=Ceil("x1") / 2,
            daily_hours=Coalesce(
                Subquery(
                    DefaultDailyPeriod.objects.filter(calendar=OuterRef("calendar"))
                    .values("calendar")
                    .annotate(s=Sum("total_hours"))
                    .values("s")[:1]
                ),
                0.0,
            ),
            available_vacation_balance_previous_year=Coalesce(
                Subquery(
                    EmployeeYearBalance.objects.filter(employee=OuterRef("pk"), year=end_of_month.year - 1).values(
                        "actual_total_vacation_hourly_balance_in_days"
                    )[:1]
                ),
                0.0,
            ),
            available_vacation_balance_current_year=Coalesce(
                Subquery(
                    EmployeeYearBalance.objects.filter(employee=OuterRef("pk"), year=end_of_month.year).values(
                        "actual_total_vacation_hourly_balance_in_days"
                    )[:1]
                ),
                0.0,
            ),
            available_vacation_balance_next_year=Coalesce(
                Subquery(
                    EmployeeYearBalance.objects.filter(employee=OuterRef("pk"), year=end_of_month.year + 1).values(
                        "actual_total_vacation_hourly_balance_in_days"
                    )[:1]
                ),
                0.0,
            ),
            long_vacation_in_hours=Value(long_vacation_days) * F("daily_hours") * F("occupancy_rate"),
            took_long_vacations=Exists(
                AbsenceRequestPeriods.objects.filter(
                    consecutive_hours_count__gte=OuterRef("long_vacation_in_hours"),
                    request__employee__id=OuterRef("pk"),
                    request__status=AbsenceRequest.Status.APPROVED,
                    request__type__is_timeoff=True,
                    date__year=end_of_month.year,
                )
            ),
        )

    @classmethod
    def get_administrators(cls) -> QuerySet[User]:
        """
        Utility classmethod that returns the HR module administrators

        Returns:
            The administrator user accounts (as queryset)
        """
        return (
            get_user_model()
            .objects.filter(is_active=True, profile__isnull=False)
            .filter(
                Q(groups__permissions__codename="administrate_absencerequest")
                | Q(user_permissions__codename="administrate_absencerequest")
            )
            .distinct()
        )

    @classmethod
    def is_administrator(cls, user: "User") -> bool:
        """
        Check if the given user account has administrator privilege

        Args:
            user: User to check

        Returns:
            True if user is an administrator
        """
        return cls.get_administrators().filter(id=user.id).exists() or (user.is_superuser and user.is_active)

    @classmethod
    def get_employee_absence_periods_df(
        cls,
        calendar: "DayOffCalendar",
        start: date,
        end: date,
        employees: QuerySet[SelfEmployeeHumanResource],
        only_employee_with_absence_periods: bool = False,
    ) -> pd.DataFrame:
        """
        Utility function that gets the subsequent absence dataframe from absence periods, employee weekly off period and days off and concat
        a unique dataframe containing all the employees non working periods.

        Args:
            calendar: The calendar to use as base calendar
            start: The start filter
            end: The end filter
            employees: A queryset of employees to get the off periods from
            only_employee_with_absence_periods: True if we want to clean rows from employees without at least one absence request during that date range

        Returns:
                A dataframe whose columns are [employee, employee_repr, period, start, end, date, type, status]
        """
        periods_map = dict()

        def _get_timespan(period_id, val_date):
            if period_id not in periods_map:
                periods_map[period_id] = DefaultDailyPeriod.objects.get(id=period_id)
            period = periods_map[period_id]
            return (
                make_naive(period.get_lower_datetime(val_date), timezone=calendar.timezone),
                make_naive(period.get_upper_datetime(val_date), timezone=calendar.timezone),
            )

        df_periods = AbsenceRequestPeriods.get_periods_as_df(start, end, employee__in=employees)
        if only_employee_with_absence_periods:
            employees = employees.filter(id__in=df_periods.employee.unique())
        df_employee_weekly_off_periods = EmployeeWeeklyOffPeriods.get_employee_weekly_periods_df(
            start, end, employee__in=employees
        )
        df_day_offs = calendar.get_day_off_per_employee_df(start, end, employees)

        df = pd.concat(
            [df_periods, df_employee_weekly_off_periods, df_day_offs],
            axis=0,
            ignore_index=True,
        )

        df[["start", "end"]] = (
            df[["period", "date"]].apply(lambda x: _get_timespan(x["period"], x["date"]), axis=1).apply(pd.Series)
        )
        df["start"] = pd.to_datetime(df["start"])
        df["end"] = pd.to_datetime(df["end"])
        df["employee_repr"] = df.employee.map(dict(employees.values_list("id", "computed_str")))
        return df

    @classmethod
    def get_end_of_month_employee_balance_report_df(
        cls, active_employees: QuerySet[SelfEmployeeHumanResource], end_of_month: date, convert_in_days: bool = True
    ) -> pd.DataFrame:
        """
        A utility function that generate a statistics vacation usage dataframe for a list of employees.

        Args:
            active_employees: A queryset of employees
            end_of_month: The last date for periods to be used in the report
            convert_in_days: True if the resulting statistics needs to be converted into days from hours

        Returns:
            A dataframe with columns ["Employee", "Total Balance", "End of Month Usage", "Available Balance End of Month"
        """

        final_columns_name_mapping = {
            "employee": "Employee",
            "total_balance": "Total Balance",
            "current_year_usage": "End of Month Usage",
            "eod_remaining_absence_days": "Available Balance End of Month",
        }

        df_balances_current_year = (
            pd.DataFrame(
                EmployeeYearBalance.objects.get_queryset_at_date(to_date=end_of_month, today=end_of_month)
                .filter(year=end_of_month.year, employee__in=active_employees)
                .values("employee", "_balance", "_daily_hours")
            )
            .set_index("employee")
            .rename(columns={"_balance": "current_year_available_balance"})
        )
        df_balances_previous_year = (
            pd.DataFrame(
                EmployeeYearBalance.objects.get_queryset_at_date(to_date=end_of_month, today=end_of_month)
                .filter(year=end_of_month.year - 1, employee__in=active_employees)
                .values("employee", "_total_vacation_hourly_balance")
            )
            .set_index("employee")
            .rename(columns={"_total_vacation_hourly_balance": "previous_year_available_balance"})
        )

        df_usage_current_year = (
            pd.DataFrame(
                AbsenceRequestPeriods.vacation_objects.filter(
                    employee__in=active_employees, date__year=end_of_month.year, date__lte=end_of_month
                ).values("employee", "_total_hours")
            )
            .groupby("employee")
            .sum()
            .rename(columns={"_total_hours": "current_year_usage"})
        )
        df_usage_previous_year = (
            pd.DataFrame(
                EmployeeYearBalance.objects.get_queryset_at_date(
                    from_date=date(end_of_month.year, 1, 1), to_date=end_of_month, today=end_of_month
                )
                .filter(year=end_of_month.year - 1, employee__in=active_employees)
                .values("employee", "_total_vacation_hourly_usage")
            )
            .set_index("employee")
            .rename(columns={"_total_vacation_hourly_usage": "previous_year_usage"})
        )

        df = pd.concat(
            [df_balances_current_year, df_balances_previous_year, df_usage_current_year, df_usage_previous_year],
            axis=1,
        ).fillna(0)

        df["total_balance"] = (
            df.current_year_available_balance + df.previous_year_available_balance + df.previous_year_usage
        )
        df["eod_remaining_absence_days"] = df.total_balance - df.current_year_usage
        if convert_in_days:
            df[["total_balance", "current_year_usage", "eod_remaining_absence_days"]] = df[
                ["total_balance", "current_year_usage", "eod_remaining_absence_days"]
            ].divide(df["_daily_hours"], axis=0)
        df = df[df["total_balance"] != 0].reset_index()
        df.employee = df.employee.map(dict(active_employees.values_list("id", "computed_str")))

        df = df.rename(columns=final_columns_name_mapping)
        return df.drop(columns=df.columns.difference(final_columns_name_mapping.values()))


class EmployeeYearBalanceDefaultManager(models.Manager):
    def _annotate_queryset(
        self, qs, from_date: date | None = None, to_date: date | None = None, today: date | None = None
    ):
        """
        Intermediary private method to annotate the balance queryset with the usage variables

        Args:
            qs: Balance queryset
            from_date: Optional, All periods before this date will be excluded from the usage
            to_date: Optional, All periods after this date will be excluded from the usage
            today: Optional, Date at which the statisticis needs to be considered. Usefull particularly to define if the previous year balance available is usable.

        Returns:
            The annotated queryset
        """
        if not today:
            today = date.today()
        try:
            latest_date_for_last_year_vacation_availibility = get_previous_year_balance_expiration_date(today.year)
            base_periods = AbsenceRequestPeriods.vacation_objects.filter(balance=OuterRef("pk"))
            base_mandatory_days_off = DayOff.objects.filter(
                calendar=OuterRef("employee__calendar"), date__year=OuterRef("year"), count_as_holiday=False
            )
            if to_date:
                base_periods = base_periods.filter(date__lte=to_date)
                base_mandatory_days_off = base_mandatory_days_off.filter(date__lte=to_date)

            if from_date:
                base_periods = base_periods.filter(date__gte=from_date)
                base_mandatory_days_off = base_mandatory_days_off.filter(date__gte=from_date)
            return qs.annotate(
                _given_balance=ExpressionWrapper(
                    Coalesce(
                        Subquery(
                            BalanceHourlyAllowance.objects.filter(balance=OuterRef("pk"))
                            .values("balance")
                            .annotate(s=Sum("hourly_allowance"))
                            .values("s")[:1]
                        ),
                        0,
                    ),
                    output_field=FloatField(),
                ),
                _balance=F("extra_balance") + Ceil("_given_balance"),
                _daily_hours=Coalesce(
                    Subquery(
                        DefaultDailyPeriod.objects.filter(calendar=OuterRef("employee__calendar"))
                        .values("calendar")
                        .annotate(s=Sum("total_hours"))
                        .values("s")[:1]
                    ),
                    0.0,
                ),
                _number_mandatory_days_off_in_days=ExpressionWrapper(
                    Coalesce(Subquery(base_mandatory_days_off.annotate(c=Count("date__year")).values("c")[:1]), 0.0),
                    output_field=FloatField(),
                ),
                _number_mandatory_days_off=F("_number_mandatory_days_off_in_days") * F("_daily_hours"),
                _total_vacation_hourly_usage=Coalesce(
                    Subquery(base_periods.values("balance").annotate(s=Sum("_total_hours")).values("s")[:1]),
                    0.0,
                ),
                _total_vacation_hourly_balance=F("_balance")
                - F("_total_vacation_hourly_usage")
                - F("_number_mandatory_days_off"),
                today=Value(today),
                actual_total_vacation_hourly_balance=Case(
                    When(
                        Q(year__lt=today.year) & Q(today__gte=latest_date_for_last_year_vacation_availibility),
                        then=0.0,
                    ),
                    default=F("_total_vacation_hourly_balance"),
                ),
                _balance_in_days=Case(When(_daily_hours=0, then=None), default=F("_balance") / F("_daily_hours")),
                _total_vacation_hourly_usage_in_days=Case(
                    When(_daily_hours=0, then=None), default=F("_total_vacation_hourly_usage") / F("_daily_hours")
                ),
                _total_vacation_hourly_balance_in_days=Case(
                    When(_daily_hours=0, then=None), default=F("_total_vacation_hourly_balance") / F("_daily_hours")
                ),
                actual_total_vacation_hourly_balance_in_days=Case(
                    When(_daily_hours=0, then=None),
                    default=F("actual_total_vacation_hourly_balance") / F("_daily_hours"),
                ),
            )
        except ProgrammingError:
            return qs

    def get_queryset(self):
        """
        Default Manager queryset
        """
        return self._annotate_queryset(super().get_queryset())

    def get_queryset_at_date(
        self, from_date: date | None = None, to_date: date | None = None, today: date | None = None
    ):
        """
        Default Manager queryset
        """
        return self._annotate_queryset(super().get_queryset(), from_date=from_date, to_date=to_date, today=today)


class BalanceHourlyAllowance(models.Model):
    balance = models.ForeignKey(
        "wbhuman_resources.EmployeeYearBalance",
        on_delete=models.CASCADE,
        verbose_name="Balance",
        related_name="monthly_allowances",
    )
    period_index = models.PositiveIntegerField()
    hourly_allowance = models.FloatField()

    class Meta:
        verbose_name = _("Monthly Allowance")
        verbose_name_plural = _("Monthly Allowance")
        unique_together = ("balance", "period_index")
        indexes = [
            models.Index(fields=["balance"]),
            models.Index(fields=["balance", "period_index"]),
        ]


class EmployeeYearBalance(ComplexToStringMixin, models.Model):
    employee = models.ForeignKey(
        EmployeeHumanResource,
        related_name="balances",
        on_delete=models.CASCADE,
        verbose_name=_("Employee"),
        help_text=_("The employee having that year balance"),
    )
    year = YearField(verbose_name=_("Year"))
    extra_balance = models.FloatField(
        default=0,
        verbose_name=_("Extra Balance (in hours)"),
        help_text=_("The yearly extra balance (in hours)"),
    )

    objects = EmployeeYearBalanceDefaultManager()

    def compute_str(self):
        return "Balance {}: {}".format(self.year, self.employee)

    class Meta:
        verbose_name = _("Employee Year Balance")
        verbose_name_plural = _("Employee Year Balances")
        unique_together = ("employee", "year")
        indexes = [
            models.Index(fields=["employee", "year"]),
        ]

    @property
    @admin.display(description="Yearly allowance (in hours)")
    def balance(self) -> float:
        given_balance = self.monthly_allowances.aggregate(s=Sum("hourly_allowance"))["s"] or 0.0
        return getattr(self, "_balance", math.ceil(given_balance) + self.extra_balance)

    @property
    @admin.display(description="Daily hours")
    def daily_hours(self) -> float:
        return getattr(self, "_daily_hours", self.employee.calendar.get_daily_hours())

    @property
    @admin.display(description="Mandatory days off (In days)")
    def number_mandatory_days_off_in_days(self) -> float:
        return getattr(
            self,
            "_number_mandatory_days_off",
            self.employee.calendar.days_off.filter(date__year=self.year, count_as_holiday=False).count(),
        )

    @property
    @admin.display(description="Mandatory Hours off")
    def number_mandatory_days_off(self) -> float:
        return getattr(
            self, "_number_mandatory_days_off_in_days", self.number_mandatory_days_off_in_days * self.daily_hours
        )

    @property
    @admin.display(description="Total vacation hourly usage")
    def total_vacation_hourly_usage(self) -> float:
        return getattr(
            self,
            "_total_vacation_hourly_usage",
            self.periods.filter(
                request__type__is_vacation=True, request__status=AbsenceRequest.Status.APPROVED
            ).aggregate(s=Sum("_total_hours"))["s"]
            or 0.0,
        )

    @property
    @admin.display(description="Total available yearly balance")
    def total_vacation_hourly_balance(self) -> float:
        return getattr(
            self,
            "_total_vacation_hourly_balance",
            self.balance - self.total_vacation_hourly_usage - self.number_mandatory_days_off,
        )

    @property
    @admin.display(description="Yearly allowance (in days)")
    def balance_in_days(self) -> float:
        return getattr(self, "_balance_in_days", self.balance / self.daily_hours)

    @property
    @admin.display(description="Total vacation hourly usage (in days)")
    def total_vacation_hourly_usage_in_days(self) -> float:
        return getattr(
            self, "_total_vacation_hourly_usage_in_days", self.total_vacation_hourly_usage / self.daily_hours
        )

    @property
    @admin.display(description="Total available yearly balance (in days)")
    def total_vacation_hourly_balance_in_days(self) -> float:
        return getattr(
            self, "_total_vacation_hourly_balance_in_days", self.total_vacation_hourly_balance / self.daily_hours
        )

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"

    @classmethod
    def get_representation_endpoint(cls) -> str:
        return "wbhuman_resources:employeeyearbalancerepresentation-list"


class EmployeeWeeklyOffPeriods(ComplexToStringMixin, models.Model):
    employee = models.ForeignKey(
        "wbhuman_resources.EmployeeHumanResource",
        related_name="default_periods_relationships",
        on_delete=models.CASCADE,
        verbose_name=_("Employee"),
    )

    period = models.ForeignKey(
        "wbhuman_resources.DefaultDailyPeriod",
        related_name="employees_relationships",
        on_delete=models.CASCADE,
        verbose_name=_("Off Period"),
    )
    weekday = models.PositiveIntegerField(
        validators=[MinValueValidator(0), MaxValueValidator(6)]
    )  # Valid weekday from python convention (e.g. Monday is 0)

    def compute_str(self) -> str:
        return "{} off on the {} in {}".format(
            self.employee.profile.full_name, self.period.title, date(1, 1, self.weekday + 1).strftime("%A")
        )

    def get_timespan(self, val_date: date) -> TimestamptzRange:
        """
        Get the combined datetime range for this employee weekly off period given a date

        Args:
            val_date: A date

        Returns:
            The datetime range spanning this periods on that valuation date
        """
        return TimestamptzRange(
            lower=self.period.get_lower_datetime(val_date), upper=self.period.get_upper_datetime(val_date)
        )

    class Meta:
        verbose_name = _("Employee Weekly off Period")
        verbose_name_plural = _("Employee Weekly off Periods")
        unique_together = ("employee", "period", "weekday")
        indexes = [
            models.Index(fields=["employee", "period", "weekday"]),
        ]

    @classmethod
    def get_representation_endpoint(cls) -> str:
        return "wbhuman_resources:employeeweeklyoffperiodrepresentation-list"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"

    @classmethod
    def get_employee_weekly_periods_df(cls, start: date, end: date, **extra_filter_kwargs) -> pd.DataFrame:
        """
        This utility function provides a way to duplicate the employee weekly off periods into a timeserie of absence request whose type is "Day Off" and status is "APPROVED"
        Args:
            start: The beginning of the time period
            end: The end of the time period
            **extra_filter_kwargs: extra filter argument as dictionary to filter down the list of Employee weekly periods (usually to filter out employees)

        Returns:
            A dataframe whose columns is ["status", "type", "employee", "period", "date"]
        """
        df_periods_off = pd.DataFrame(
            EmployeeWeeklyOffPeriods.objects.filter(**extra_filter_kwargs).values("employee", "period", "weekday"),
            columns=["employee", "period", "weekday"],
        )
        df_periods_off = df_periods_off.merge(
            pd.date_range(start, end, freq="W").to_series(name="sunday"), how="cross"
        )
        df_periods_off["date"] = df_periods_off["sunday"] - pd.TimedeltaIndex(6 - df_periods_off["weekday"], unit="D")
        df_periods_off.date = df_periods_off.date.dt.date
        del df_periods_off["weekday"]
        del df_periods_off["sunday"]
        df_periods_off["status"] = AbsenceRequest.Status.APPROVED.value
        df_periods_off["type"] = "Day Off"
        return df_periods_off


class Position(ComplexToStringMixin, WBModel, MPTTModel):
    color = ColorField(default="#FF0000")
    name = models.CharField(max_length=256)
    parent = TreeForeignKey(
        "wbhuman_resources.Position",
        related_name="children",
        null=True,
        blank=True,
        on_delete=models.CASCADE,
        verbose_name=_("Parent Positions"),
    )
    groups = models.ManyToManyField(Group, related_name="human_resources_positions", blank=True)
    height = models.IntegerField(default=0)
    manager = models.ForeignKey(
        "directory.Person",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        verbose_name=_("Department Manager"),
        related_name="managed_positions",
    )

    class Meta:
        verbose_name = _("Position")
        verbose_name_plural = _("Positions")

    def compute_str(self) -> str:
        if self.parent and self.parent.computed_str:
            return f"{self.name} ({self.parent.computed_str})"
        return f"{self.name}"

    def save(self, *args, **kwargs):
        self.computed_str = self.compute_str()
        super().save(*args, **kwargs)

    def get_employees(self) -> QuerySet[EmployeeHumanResource]:
        """
        Get the queryset of employees that are within this position hierarchy

        Returns:
            A queryset of employees
        """
        return EmployeeHumanResource.objects.filter(
            is_active=True, position__in=self.get_descendants(include_self=True)
        )

    # @classmethod
    # def to_graph(cls):
    #     nr_vertices = 25
    #     v_label = list(map(str, range(nr_vertices)))
    #     G = Graph.Tree(nr_vertices, 2)  # 2 stands for children number
    #     lay = G.layout('rt')
    #
    #     position = {k: lay[k] for k in range(nr_vertices)}
    #     Y = [lay[k][1] for k in range(nr_vertices)]
    #     M = max(Y)
    #
    #     es = EdgeSeq(G)  # sequence of edges
    #     E = [e.tuple for e in G.es]  # list of edges
    #
    #     L = len(position)
    #     Xn = [position[k][0] for k in range(L)]
    #     Yn = [2 * M - position[k][1] for k in range(L)]
    #     Xe = []
    #     Ye = []
    #     for edge in E:
    #         Xe += [position[edge[0]][0], position[edge[1]][0], None]
    #         Ye += [2 * M - position[edge[0]][1], 2 * M - position[edge[1]][1], None]
    #
    #     labels = v_label
    @classmethod
    def get_endpoint_basename(cls) -> str:
        return "wbhuman_resources:position"

    @classmethod
    def get_representation_endpoint(cls) -> str:
        return "wbhuman_resources:positionrepresentation-list"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"

    @classmethod
    def get_representation_label_key(cls) -> str:
        return "{{ computed_str }}"


@receiver(post_save, sender=EmployeeHumanResource)
def post_save_employee(sender, instance, created, **kwargs):
    """
    Post save signal, Ensure that employee is among the employees of the CRM company profile upon save
    """
    if created:
        if not instance.weekly_off_periods.exists():
            # Default in creating the weekend default period off
            for default_period in instance.calendar.default_periods.all():
                EmployeeWeeklyOffPeriods.objects.get_or_create(
                    employee=instance, period=default_period, weekday=5
                )  # Saturday
                EmployeeWeeklyOffPeriods.objects.get_or_create(
                    employee=instance, period=default_period, weekday=6
                )  # Sunday
        [start, end] = EmployeeHumanResource.ExtraDaysBalanceFrequency[instance.extra_days_frequency].get_date_range(
            instance.enrollment_at
        )
        instance.assign_vacation_allowance_from_range(start.date(), end.date())

        try:
            main_company = Company.objects.get(id=get_main_company())
            rel = EmployerEmployeeRelationship.objects.get_or_create(
                employee=instance.profile,
                employer=main_company,
                defaults={
                    "primary": True,
                    "position": (
                        CRMPosition.objects.get_or_create(title=instance.position.name)[0]
                        if instance.position
                        else None
                    ),
                },
            )[0]
            rel.primary = True
            rel.save()
        except Company.DoesNotExist:
            pass

    # We assign or unassign the auth.Group position based on the active status
    if instance.is_active:
        instance.assign_position_groups()
    else:
        instance.unassign_position_groups()


@receiver(post_delete, sender=Position)
def post_delete_position(sender, instance, **kwargs):
    """
    Post delete Position logic
    """
    # We compute the height given the level of this position and its leaf node level
    if leaf_position := instance.get_family().filter(children__isnull=True).order_by("-level").first():
        instance.height = leaf_position.level - instance.level

    # We recompute the height for all ancestors if this position is a leaf node
    if instance.height == 0:
        for pos in instance.get_ancestors():
            pos.save()


@receiver(m2m_changed, sender=Position.groups.through)
def position_groups_to_employee(
    sender: Type[Position.groups.through], instance: Position, action: str, pk_set: set[int], **kwargs
):
    employees = EmployeeHumanResource.objects.filter(
        profile__isnull=False,
        profile__user_account__isnull=False,
        position__in=instance.get_descendants(include_self=True),
    )
    groups_to_changed = Group.objects.filter(id__in=pk_set)
    for employee in employees:
        user = employee.profile.user_account
        for group in groups_to_changed:
            if action == "post_add":
                if group not in user.groups.all():
                    user.groups.add(group)
            elif action in ["post_remove", "post_clear"]:
                user.groups.remove(group)


@shared_task
def deactivate_profile_as_task(requester_id: int, employee_id: int, substitute_id: Optional[int] = None):
    """
    Call the deactivation method as a async task

    Args:
        requester_id: The User account id of the user asking for the deactivation
        employee_id: The deactivated user profile
        substitute_id: The potential Profile id to replace this employee's resources

    """
    requester = get_user_model().objects.get(id=requester_id)
    employee = EmployeeHumanResource.objects.get(id=employee_id)
    substitute = None
    if substitute_id:
        substitute = Person.objects.get(id=substitute_id)
    messages = employee.deactivate(substitute)

    send_notification(
        code="wbhuman_resources.employeehumanresource.deactivate",
        title=gettext("{employee} deactivation ended and was successful").format(
            employee=employee.profile.computed_str
        ),
        body=gettext("The following actions have been done: \n{messages}").format(messages=messages),
        user=requester,
    )



from django.conf import settings
from django.forms import ValidationError
from django.utils.translation import gettext as _
from dynamic_preferences.preferences import Section
from dynamic_preferences.registries import global_preferences_registry
from dynamic_preferences.types import (
    BooleanPreference,
    IntegerPreference,
    LongStringPreference,
    ModelChoicePreference,
    StringPreference,
)

from wbhuman_resources.models.calendars import DayOffCalendar

human_resources = Section("wbhuman_resources")

# Employee dynamic preferences


@global_preferences_registry.register
class EmployeeDefaultCalendarEntry(ModelChoicePreference):
    section = human_resources
    name = "employee_default_calendar"
    queryset = DayOffCalendar.objects.all()
    default = None


@global_preferences_registry.register
class DefaultVacationDaysPreference(IntegerPreference):
    section = human_resources
    name = "default_vacation_days"
    default = 25

    verbose_name = _("Default employee vacation days")
    help_text = _("The number of vacation days allocated to an employee working full time")

    def validate(self, value):
        if not isinstance(value, int) or value <= 1:
            raise ValidationError(_("Only positive natural numbers allowed."))


@global_preferences_registry.register
class DefaultFromEmailAddressPreference(StringPreference):
    section = human_resources
    name = "default_from_email_address"
    default = settings.DEFAULT_FROM_EMAIL

    verbose_name = "The default from email address"
    help_text = "The default from email address used to send hr related emails"


@global_preferences_registry.register
class NumberOfMonthsBeforeBalanceExpiration(IntegerPreference):
    section = human_resources
    name = "number_of_month_before_balance_expiration"
    default = 12 * 5  # Default to 5 years
    verbose_name = _("The number of month before yearly balance expiration")
    help_text = _(
        "The number of months before any yearly balance is considered expired. The count starts the first day of the next balance year"
    )


@global_preferences_registry.register
class LongVacationNumberOfDaysPreference(IntegerPreference):
    section = human_resources
    name = "long_vacation_number_of_days"
    default = 10
    help_text = _("The number of days after which a vacation is considered a long vacation.")
    verbose_name = _("Long Vacation Number of days")


# Monthly report accounting preference


@global_preferences_registry.register
class AccountingCompanyEmails(LongStringPreference):
    section = human_resources
    name = "accounting_company_emails"
    default = "no-reply@stainly-bench.com;"
    verbose_name = _("Accounting email destination")
    help_text = _("The accounting company emails to send automatic reports to, as a comma separated list")


# Calendar preferences


@global_preferences_registry.register
class CalendarDefaultPublicHolidayPackagePreference(StringPreference):
    section = human_resources
    name = "calendar_default_public_holiday_package"
    default = "europe.Switzerland"

    verbose_name = _("Default calendar Public holiday package")
    help_text = _("Package where the correct public holidays are found. Defaults to Europe Switzerland")

    def validate(self, value):
        if len(value.split(".")) != 2:
            raise ValidationError(_("The preference has to be in the format of <continent>.<region>"))


@global_preferences_registry.register
class CalendarDefaultTimezonePreference(StringPreference):
    section = human_resources
    name = "calendar_default_timezone"
    default = "UTC"

    verbose_name = _("Default calendar timezone")
    help_text = _("The default calendar timezone")


@global_preferences_registry.register
class AreExternalEmployeesConsideredAsInternal(BooleanPreference):
    section = human_resources
    name = "is_external_considered_as_internal"
    default = False

    verbose_name = _("Are external employee considered as internal?")
    help_text = _("If True, will consider the external employee (and any related logic) as internal employee")



import zoneinfo
from datetime import date, datetime, time, timedelta

import pandas as pd
import pytest
from django.contrib.auth.models import Group, Permission
from django.contrib.contenttypes.models import ContentType
from django.db import models
from dynamic_preferences.registries import global_preferences_registry
from faker import Faker
from psycopg.types.range import TimestamptzRange
from wbcore.contrib.directory.models import EmployerEmployeeRelationship

from wbhuman_resources.factories import (
    DayOffCalendarFactory,
    DayOffFactory,
    EmployeeYearBalanceFactory,
    TimeOffRequestFactory,
    VacationRequestFactory,
)
from wbhuman_resources.models.employee import (
    EmployeeHumanResource,
    EmployeeWeeklyOffPeriods,
    EmployeeYearBalance,
    post_save_employee,
)
from wbhuman_resources.models.preferences import (
    get_previous_year_balance_expiration_date,
)

this_year_minimum_date_to_consume_previous_year_balance = get_previous_year_balance_expiration_date(date.today().year)


fake = Faker()


@pytest.mark.django_db
class TestEmployeeHumanResource:
    # Permission base tests
    def test_get_administrators(self, employee_human_resource, user_factory):
        user_factory.create(is_superuser=True)  # superuser
        admin = user_factory.create()
        admin.user_permissions.add(Permission.objects.get(codename="administrate_absencerequest"))
        user_factory.create()  # normal_user
        assert set(EmployeeHumanResource.get_administrators()) == {admin}

    def test_is_administrator_super_user(self, user_factory):
        superuser = user_factory.create(is_superuser=True)
        assert EmployeeHumanResource.is_administrator(superuser)

    def test_is_administrator_normal_user(self, user):
        assert not EmployeeHumanResource.is_administrator(user)

    def test_is_administrator_user_with_permission(self, user_factory):
        admin = user_factory.create()
        admin.user_permissions.add(Permission.objects.get(codename="administrate_absencerequest"))
        assert EmployeeHumanResource.is_administrator(admin)

    def test_get_managed_employees(self, employee_human_resource_factory, position_factory):
        top_manager = employee_human_resource_factory.create()
        manager1 = employee_human_resource_factory.create()
        manager2 = employee_human_resource_factory.create()

        top_position = position_factory.create(manager=top_manager.profile)
        pos_1 = position_factory.create(parent=top_position, manager=manager1.profile)
        pos_2 = position_factory.create(parent=top_position, manager=manager2.profile)

        top_manager.position = top_position
        top_manager.save()
        manager1.position = pos_1
        manager1.save()
        manager2.position = pos_2
        manager2.save()

        top_employee = employee_human_resource_factory.create(position=top_position)
        employee_1 = employee_human_resource_factory.create(position=pos_1)
        employee_2 = employee_human_resource_factory.create(position=pos_2)

        employee_without_pos = employee_human_resource_factory.create(direct_manager=manager1.profile)
        top_manager_employees = {top_manager, manager1, manager2, top_employee, employee_1, employee_2}
        assert set(top_manager.get_managed_employees(include_self=False)) == {
            manager1,
            manager2,
            top_employee,
            employee_1,
            employee_2,
        }
        assert set(top_manager.get_managed_employees(include_self=True)) == top_manager_employees
        for managee in top_manager_employees:
            assert top_manager.is_manager_of(managee, include_self=True)

        manager1_employees = {employee_1, manager1, employee_without_pos}
        assert set(manager1.get_managed_employees(include_self=True)) == manager1_employees
        assert set(manager1.get_managed_employees(include_self=False)) == {employee_1, employee_without_pos}
        for managee in manager1_employees:
            assert manager1.is_manager_of(managee, include_self=True)

        manager2_employees = {employee_2, manager2}
        assert set(manager2.get_managed_employees(include_self=True)) == manager2_employees
        assert set(manager2.get_managed_employees(include_self=False)) == {employee_2}
        for managee in manager2_employees:
            assert manager2.is_manager_of(managee, include_self=True)

        assert not employee_1.get_managed_employees(include_self=False).exists()
        assert not employee_2.get_managed_employees(include_self=False).exists()
        assert not employee_without_pos.get_managed_employees(include_self=False).exists()

    def test_get_managers(self, position_factory, employee_human_resource_factory):
        position_manager = employee_human_resource_factory.create()
        main_manager = employee_human_resource_factory.create()

        position = position_factory.create(manager=position_manager.profile)
        assert (
            next(
                employee_human_resource_factory.create(
                    direct_manager=main_manager.profile, position=position
                ).get_managers()
            )
            == main_manager.profile
        )
        assert (
            next(employee_human_resource_factory.create(direct_manager=None, position=position).get_managers())
            == position_manager.profile
        )
        with pytest.raises(StopIteration):
            next(employee_human_resource_factory.create(direct_manager=None, position=None).get_managers())

    def test_get_managers_with_global_manager(
        self, authenticated_person_factory, position_factory, employee_human_resource_factory
    ):
        global_manager_permission = Permission.objects.get(
            codename="administrate_employeehumanresource",
            content_type=ContentType.objects.get_for_model(EmployeeHumanResource),
        )
        administrate_employeehumanresource = authenticated_person_factory.create()
        administrate_employeehumanresource.user_account.user_permissions.add(global_manager_permission)
        assert (
            next(employee_human_resource_factory.create(direct_manager=None, position=None).get_managers())
            == administrate_employeehumanresource
        )

    @pytest.mark.parametrize("val_date", [(fake.date_this_decade())])
    def test_deactivate(self, employee_human_resource_factory, val_date):
        employee = employee_human_resource_factory.create(enrollment_at=val_date)
        employee.deactivate(disenrollment_date=date(val_date.year, 12, 31))
        assert not employee.is_active
        assert not employee.profile.user_account.is_active
        balance = employee.balances.get(year=val_date.year)
        assert balance.monthly_allowances.count() == 1

    def test_assign_unassign_position_groups(self, employee_human_resource):
        assert not employee_human_resource.profile.user_account.groups.exists()
        employee_human_resource.assign_position_groups()
        assert employee_human_resource.profile.user_account.groups.filter(
            id=employee_human_resource.position.groups.first().id
        ).exists()
        employee_human_resource.unassign_position_groups()
        assert not employee_human_resource.profile.user_account.groups.exists()

    # Balance & Usage tests
    def test_post_save_employee(self, employee_human_resource, company):
        # Set main CRM company
        global_preferences_registry.manager()["directory__main_company"] = company.id
        global_preferences_registry.manager().cache.clear()

        post_save_employee("SENDER ", employee_human_resource, created=True)

        # Expect the weekend days to be labeled as off period by default
        assert EmployeeWeeklyOffPeriods.objects.filter(employee=employee_human_resource, weekday=5).count() == 2
        assert EmployeeWeeklyOffPeriods.objects.filter(employee=employee_human_resource, weekday=6).count() == 2

        # Expect the employee to have balance and monthly allowance for at its enrolment year
        balance = employee_human_resource.balances.get(year=employee_human_resource.enrollment_at.year)
        assert balance.monthly_allowances.count() == 1

        # Expect the CRM position to reflect the
        assert EmployerEmployeeRelationship.objects.filter(
            employee=employee_human_resource.profile,
            employer=company,
            primary=True,
            position__title=employee_human_resource.position.name,
        ).exists()
        assert employee_human_resource.profile.user_account.groups.filter(
            id=employee_human_resource.position.groups.first().id
        ).exists()

    @pytest.mark.parametrize("start_date", [(fake.date_object())])
    def test_extract_workable_periods_basic_usage(
        self, employee_human_resource, start_date, day_off_factory, employee_weekly_off_periods_factory
    ):
        end_date = start_date + timedelta(days=7)
        start_datetime = datetime.combine(start_date, time(0, 0, 0)).astimezone(zoneinfo.ZoneInfo("UTC"))
        end_datetime = datetime.combine(end_date, time(23, 59, 59)).astimezone(zoneinfo.ZoneInfo("UTC"))
        day_off = day_off_factory.create(
            calendar=employee_human_resource.calendar, date=fake.date_between(start_date, end_date)
        )
        period_off = employee_weekly_off_periods_factory.create(employee=employee_human_resource)
        res = set(employee_human_resource.extract_workable_periods(start_datetime, end_datetime))
        expected_res = set()
        for i in range(8):
            _d = start_date + timedelta(days=i)
            if day_off.date != _d:
                for period in employee_human_resource.calendar.default_periods.all():
                    if period_off.period != period or period_off.weekday != _d.weekday():
                        expected_res.add((_d, period))
        assert expected_res == res

    def test_assign_vacation_allowance_from_range(self, employee_human_resource):
        pass

    @pytest.mark.parametrize("year_str", [(fake.year())])
    def test_get_or_create_balance(self, employee_human_resource, year_str):
        assert employee_human_resource.get_or_create_balance(int(year_str))[0].year == int(year_str)


@pytest.mark.django_db
class TestEmployeeYearBalance:
    @pytest.fixture
    def test_args(self, start_date):
        balance = EmployeeYearBalanceFactory.create(year=start_date.year)
        default_period1 = balance.employee.calendar.default_periods.earliest("lower_time")
        default_period2 = balance.employee.calendar.default_periods.latest("lower_time")

        start_r2 = fake.date_between_dates(start_date, date(balance.year, 12, 31))
        start_r3 = fake.date_between_dates(date(balance.year, 1, 1), start_date)
        vacation_request1 = VacationRequestFactory.create(
            employee=balance.employee,
            period=TimestamptzRange(
                default_period1.get_lower_datetime(start_date),
                default_period2.get_upper_datetime(start_date) + timedelta(days=2),
            ),
        )
        vacation_request2 = VacationRequestFactory.create(
            employee=balance.employee,
            period=TimestamptzRange(
                default_period1.get_lower_datetime(start_r2),
                default_period2.get_upper_datetime(start_r2) + timedelta(days=2),
            ),
        )
        timeoff_request = TimeOffRequestFactory.create(
            employee=balance.employee,
            period=TimestamptzRange(
                default_period1.get_lower_datetime(start_r3),
                default_period2.get_upper_datetime(start_r3) + timedelta(days=2),
            ),
        )

        mandatatory_day_off = DayOffFactory.create(
            calendar=balance.employee.calendar,
            count_as_holiday=False,
            date=fake.date_between_dates(date(balance.year, 1, 5), date(balance.year, 12, 31)),
        )
        DayOffFactory.create(
            calendar=DayOffCalendarFactory.create(title="other calendar", resource="other.resource"),
            count_as_holiday=False,
            date=fake.date_between_dates(date(balance.year, 1, 5), date(balance.year, 12, 31)),
        )  # Other calendar
        return (
            balance,
            default_period1,
            default_period2,
            vacation_request1,
            vacation_request2,
            timeoff_request,
            mandatatory_day_off,
        )

    # test properties
    @pytest.mark.parametrize("start_date", [(fake.date_this_year())])
    def test_balance(self, test_args, start_date):
        [
            balance,
            default_period1,
            default_period2,
            vacation_request1,
            vacation_request2,
            timeoff_request,
            mandatatory_day_off,
        ] = test_args
        expected_res = (
            balance.extra_balance + balance.monthly_allowances.aggregate(s=models.Sum("hourly_allowance"))["s"]
        )
        assert balance.balance == expected_res
        assert EmployeeYearBalance.objects.get(id=balance.id)._balance == expected_res

    @pytest.mark.parametrize("start_date", [(fake.date_this_year())])
    def test_daily_hours(self, test_args, start_date):
        [
            balance,
            default_period1,
            default_period2,
            vacation_request1,
            vacation_request2,
            timeoff_request,
            mandatatory_day_off,
        ] = test_args

        expected_res = default_period1.total_hours + default_period2.total_hours
        assert balance.daily_hours == expected_res
        assert EmployeeYearBalance.objects.get(id=balance.id)._daily_hours == expected_res

    @pytest.mark.parametrize("start_date", [(fake.date_this_year())])
    def test_number_mandatory_days_off_in_days(self, test_args, start_date):
        [
            balance,
            default_period1,
            default_period2,
            vacation_request1,
            vacation_request2,
            timeoff_request,
            mandatatory_day_off,
        ] = test_args
        expected_res = 1
        assert balance.number_mandatory_days_off_in_days == expected_res
        assert EmployeeYearBalance.objects.get(id=balance.id)._number_mandatory_days_off_in_days == expected_res

    @pytest.mark.parametrize("start_date", [(fake.date_this_year())])
    def test_number_mandatory_days_off(self, test_args, start_date):
        [
            balance,
            default_period1,
            default_period2,
            vacation_request1,
            vacation_request2,
            timeoff_request,
            mandatatory_day_off,
        ] = test_args
        expected_res = balance.employee.calendar.get_daily_hours()
        assert balance.number_mandatory_days_off == expected_res
        assert EmployeeYearBalance.objects.get(id=balance.id)._number_mandatory_days_off == expected_res

    @pytest.mark.parametrize("start_date", [(fake.date_this_year())])
    def test_total_vacation_hourly_usage(self, test_args, start_date):
        [
            balance,
            default_period1,
            default_period2,
            vacation_request1,
            vacation_request2,
            timeoff_request,
            mandatatory_day_off,
        ] = test_args
        expected_res = vacation_request1.total_hours + vacation_request2.total_hours
        assert balance.total_vacation_hourly_usage == expected_res
        assert EmployeeYearBalance.objects.get(id=balance.id)._total_vacation_hourly_usage == expected_res

    @pytest.mark.parametrize("start_date", [(fake.date_this_year())])
    def test_total_vacation_hourly_balance(self, test_args, start_date):
        [
            balance,
            default_period1,
            default_period2,
            vacation_request1,
            vacation_request2,
            timeoff_request,
            mandatatory_day_off,
        ] = test_args

        expected_res = (
            balance.extra_balance
            + balance.monthly_allowances.aggregate(s=models.Sum("hourly_allowance"))["s"]
            - vacation_request1.total_hours
            - vacation_request2.total_hours
            - balance.employee.calendar.get_daily_hours()
        )
        assert balance.total_vacation_hourly_balance == expected_res
        assert EmployeeYearBalance.objects.get(id=balance.id)._total_vacation_hourly_balance == expected_res

    @pytest.mark.parametrize("start_date", [(fake.date_this_year())])
    def test_balance_in_days(self, test_args, start_date):
        [
            balance,
            default_period1,
            default_period2,
            vacation_request1,
            vacation_request2,
            timeoff_request,
            mandatatory_day_off,
        ] = test_args
        expected_res = (
            balance.extra_balance + balance.monthly_allowances.aggregate(s=models.Sum("hourly_allowance"))["s"]
        ) / balance.employee.calendar.get_daily_hours()
        assert balance.balance_in_days == expected_res
        assert EmployeeYearBalance.objects.get(id=balance.id)._balance_in_days == expected_res

    @pytest.mark.parametrize("start_date", [(fake.date_this_year())])
    def test_total_vacation_hourly_usage_in_days(self, test_args, start_date):
        [
            balance,
            default_period1,
            default_period2,
            vacation_request1,
            vacation_request2,
            timeoff_request,
            mandatatory_day_off,
        ] = test_args
        expected_res = (
            vacation_request1.total_hours + vacation_request2.total_hours
        ) / balance.employee.calendar.get_daily_hours()
        assert balance.total_vacation_hourly_usage_in_days == expected_res
        assert EmployeeYearBalance.objects.get(id=balance.id)._total_vacation_hourly_usage_in_days == expected_res

    @pytest.mark.parametrize("start_date", [(fake.date_this_year())])
    def test_total_vacation_hourly_balance_in_days(self, test_args, start_date):
        [
            balance,
            default_period1,
            default_period2,
            vacation_request1,
            vacation_request2,
            timeoff_request,
            mandatatory_day_off,
        ] = test_args
        expected_res = (
            balance.extra_balance
            + balance.monthly_allowances.aggregate(s=models.Sum("hourly_allowance"))["s"]
            - vacation_request1.total_hours
            - vacation_request2.total_hours
            - balance.employee.calendar.get_daily_hours()
        ) / balance.employee.calendar.get_daily_hours()

        assert balance.total_vacation_hourly_balance_in_days == expected_res
        assert EmployeeYearBalance.objects.get(id=balance.id)._total_vacation_hourly_balance_in_days == expected_res


@pytest.mark.django_db
class TestEmployeeWeeklyOffPeriods:
    @pytest.mark.parametrize("test_date", [(fake.date_this_year())])
    def get_timespan(self, employee_weekly_off_periods, test_date):
        timespan = employee_weekly_off_periods.get_timespan(test_date)
        assert timespan.lower == employee_weekly_off_periods.period.get_lower_datetime(test_date)
        assert timespan.upper == employee_weekly_off_periods.period.get_upper_datetime(test_date)

    @pytest.mark.parametrize("test_date", [(fake.date_this_year())])
    def test_get_employee_weekly_periods_df(
        self, day_off_calendar, employee_human_resource_factory, employee_weekly_off_periods_factory, test_date
    ):
        test_date = (test_date - pd.tseries.offsets.Week(weekday=0)).date()
        employee1 = employee_human_resource_factory.create(calendar=day_off_calendar, is_active=True)
        employee2 = employee_human_resource_factory.create(calendar=day_off_calendar, is_active=True)
        employee3 = employee_human_resource_factory.create(
            calendar=day_off_calendar, is_active=False
        )  # Expect this employee to not be present

        period1 = employee_weekly_off_periods_factory.create(employee=employee1)
        period2 = employee_weekly_off_periods_factory.create(employee=employee2)
        period3 = employee_weekly_off_periods_factory.create(employee=employee3)
        res = EmployeeWeeklyOffPeriods.get_employee_weekly_periods_df(
            test_date, test_date + timedelta(days=7), employee__is_active=True
        )
        assert res.shape == (2, 5)
        res = res.set_index(["employee", "period", "date"])
        assert res.loc[
            (employee1.id, period1.period.id, test_date + timedelta(days=period1.weekday)), :
        ].values.tolist() == ["APPROVED", "Day Off"]
        assert res.loc[
            (employee2.id, period2.period.id, test_date + timedelta(days=period2.weekday)), :
        ].values.tolist() == ["APPROVED", "Day Off"]
        with pytest.raises(KeyError):
            assert res.loc[(employee3.id, period3.period.id, test_date + timedelta(days=period3.weekday)), :]


@pytest.mark.django_db
class TestPosition:
    def test_change_group_in_position_change_user_groups(self, employee_human_resource):
        employee_human_resource.assign_position_groups()
        position = employee_human_resource.position
        first_group = position.groups.first()
        second_group = Group.objects.create(name="second group")

        assert set(employee_human_resource.profile.user_account.groups.all()) == {first_group}

        position.groups.add(second_group)
        assert set(employee_human_resource.profile.user_account.groups.all()) == {first_group, second_group}

        position.groups.remove(second_group)
        assert set(employee_human_resource.profile.user_account.groups.all()) == {first_group}

        position.groups.remove(first_group)
        assert set(employee_human_resource.profile.user_account.groups.all()) == set()

    def test_get_employees(self, position_factory, employee_human_resource_factory):
        root_1 = position_factory.create()
        e_root_1 = employee_human_resource_factory.create(position=root_1)
        child_1 = position_factory.create(parent=root_1)
        e_child_1 = employee_human_resource_factory.create(position=child_1)
        employee_human_resource_factory.create(
            position=child_1, is_active=False
        )  # Unactive employee, shouldn't show in the method

        root_2 = position_factory.create()
        e_root_2 = employee_human_resource_factory.create(position=root_2)
        child_2 = position_factory.create(parent=root_2)
        e_child_2 = employee_human_resource_factory.create(position=child_2)
        employee_human_resource_factory.create(
            position=child_2, is_active=False
        )  # Unactive employee, shouldn't show in the method

        assert {*root_1.get_employees().values_list("id", flat=True)} == {e_root_1.id, e_child_1.id}
        assert {*child_1.get_employees().values_list("id", flat=True)} == {e_child_1.id}
        assert {*root_2.get_employees().values_list("id", flat=True)} == {e_root_2.id, e_child_2.id}
        assert {*child_2.get_employees().values_list("id", flat=True)} == {e_child_2.id}



from datetime import timedelta

import pandas as pd
from django.contrib import admin
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group
from django.contrib.postgres.constraints import ExclusionConstraint
from django.contrib.postgres.fields import DateTimeRangeField, RangeOperators
from django.core.exceptions import ValidationError
from django.db import models
from django.db.models.functions import Coalesce
from django.db.models.signals import post_save
from django.db.utils import ProgrammingError
from django.dispatch import receiver
from django.utils import timezone
from django.utils.functional import cached_property
from django.utils.translation import gettext, pgettext_lazy
from django.utils.translation import gettext_lazy as _
from django_fsm import FSMField, transition
from pandas._libs.tslibs.offsets import BDay
from psycopg.types.range import TimestamptzRange
from wbcore.contrib.agenda.models import CalendarItem
from wbcore.contrib.geography.models import Geography
from wbcore.contrib.icons import WBIcon
from wbcore.contrib.notifications.dispatch import send_notification
from wbcore.contrib.notifications.utils import create_notification_type
from wbcore.enums import RequestType
from wbcore.metadata.configs.buttons import ActionButton, ButtonDefaultColor
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    create_simple_display,
)
from wbcore.models import WBModel
from wbcore.utils.models import CalendarItemTypeMixin

from .calendars import DayOff, DefaultDailyPeriod
from .preferences import get_previous_year_balance_expiration_date

User = get_user_model()


def can_cancel_request(instance: "AbsenceRequest", user: "User") -> bool:
    """
    Check if the given user has cancel ability on the given request

    Args:
        instance: The request
        user: User to check permission

    Returns:
        True if the user got the right to cancel this request
    """
    if instance.period.upper < timezone.now():
        return False
    return (
        user.profile == instance.employee.profile
        or user.has_perm("wbhuman_resources.administrate_absencerequest")
        or user.profile.human_resources.is_manager_of(instance.employee)
    )


def can_validate_or_deny_request(instance: "AbsenceRequest", user: "User") -> bool:
    """
    Check if the given user can validate or deny the given request

    Args:
        instance: The request
        user: User to check permission

    Returns:
        True if the user got the right to validate or deny this request
    """
    if user.has_perm("wbhuman_resources.administrate_absencerequest"):
        return True
    elif (
        user.profile
        and (employee := getattr(user.profile, "human_resources", None))
        and (requester := instance.employee)
    ):
        return employee.is_manager_of(requester)
    return False


class AbsenceRequestType(CalendarItemTypeMixin, WBModel):
    title = models.CharField(max_length=255, verbose_name=_("Title"))

    is_vacation = models.BooleanField(
        default=False,
        verbose_name=_("Vacation"),
        help_text=_("If true, the days will be counted towards the employee's vacation balance"),
    )
    is_timeoff = models.BooleanField(
        default=False, verbose_name=_("Time-Off"), help_text=_("If true, the employee is considered as not working")
    )
    is_extensible = models.BooleanField(
        default=False,
        verbose_name=_("Extensible"),
        help_text=_("If true, allow the associated request to be extended"),
    )
    auto_approve = models.BooleanField(default=False, verbose_name=_("Auto Approve"))
    days_in_advance = models.PositiveIntegerField(default=0, verbose_name=_("Days In Advance"))

    is_country_necessary = models.BooleanField(default=False, verbose_name=_("Is country necessary"))
    crossborder_countries = models.ManyToManyField(
        to="geography.Geography",
        limit_choices_to={"level": 1},
        blank=True,
        verbose_name=_("Countries"),
        help_text=_("List of countries where crossborder activity is allowed"),
    )
    extra_notify_groups = models.ManyToManyField(
        to=Group, blank=True, related_name="notified_absence_request_types", verbose_name=_("Extra Notify Groups")
    )

    def validate_country(self, country):
        if self.is_country_necessary:
            if not country:
                raise ValueError(_("A country is necessary for this absence request type"))
            elif not self.crossborder_countries.filter(id=country.id).exists():
                raise ValueError(
                    _(
                        "You are not allowed to have crossborder activities in the specified country for the specified absence request type"
                    )
                )
        return True

    class Meta:
        verbose_name = _("Absence Request Type")
        verbose_name_plural = _("Absence Request Types")

    def __str__(self):
        return self.title

    @classmethod
    def get_choices(cls) -> list[tuple[int, str]]:
        """
        Utility method that returns all possible absence request type choices as a text choices datastructure with id as name and title as label

        We expect runtime error at runtime on non-initialized database that will be caught and an empty list will be returned in that case.

        Returns:
            A list of tuple choices
        """
        try:
            return [(absence_type.id, absence_type.title) for absence_type in cls.objects.all()]
        except (RuntimeError, ProgrammingError):
            return []

    @classmethod
    def get_endpoint_basename(cls) -> str:
        return "wbhuman_resources:absencerequesttype"

    @classmethod
    def get_representation_endpoint(cls) -> str:
        return "wbhuman_resources:absencerequesttyperepresentation-list"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"

    @classmethod
    def get_representation_label_key(cls) -> str:
        return "{{title}}"


class AbsenceRequestManager(models.Manager):
    def get_queryset(self) -> "models.QuerySet[AbsenceRequest]":
        """
        Default manager that provide a set of annotated variables for convenience
        """
        return (
            super()
            .get_queryset()
            .annotate(
                daily_hours=Coalesce(
                    models.Subquery(
                        DefaultDailyPeriod.objects.filter(calendar=models.OuterRef("employee__calendar"))
                        .values("calendar")
                        .annotate(s=models.Sum("total_hours"))
                        .values("s")[:1]
                    ),
                    0.0,
                ),
                _total_hours=Coalesce(
                    models.Subquery(
                        AbsenceRequestPeriods.objects.filter(request=models.OuterRef("pk"))
                        .values("request")
                        .annotate(s=models.Sum("_total_hours"))
                        .values("s")[:1]
                    ),
                    0.0,
                ),
                _total_vacation_hours=Coalesce(
                    models.Subquery(
                        AbsenceRequestPeriods.vacation_objects.filter(request=models.OuterRef("pk"))
                        .values("request")
                        .annotate(s=models.Sum("_total_hours"))
                        .values("s")[:1]
                    ),
                    0.0,
                ),
                _total_hours_in_days=models.F("_total_hours") / models.F("daily_hours"),
                _total_vacation_hours_in_days=models.F("_total_vacation_hours") / models.F("daily_hours"),
            )
        )


class AbsenceRequest(CalendarItem):
    """
    Stores a single Absence Request entry, related to :model:`wbhuman_resources.EmployeeHumanResource` and :model:`wbcrm.Activity`.
    """

    class Status(models.TextChoices):
        DRAFT = "DRAFT", _("Draft")
        PENDING = "PENDING", _("Pending")
        APPROVED = "APPROVED", _("Approved")
        DENIED = "DENIED", _("Denied")
        CANCELLED = "CANCELLED", _("Cancelled")

    status = FSMField(
        default=Status.DRAFT,
        choices=Status.choices,
        verbose_name=_("Status"),
        help_text=_("The request status (defaults to draft)"),
    )

    type = models.ForeignKey(
        to="wbhuman_resources.AbsenceRequestType",
        related_name="request",
        verbose_name=_("Type"),
        on_delete=models.PROTECT,
    )

    employee = models.ForeignKey(
        "wbhuman_resources.EmployeeHumanResource",
        related_name="requests",
        on_delete=models.CASCADE,
        verbose_name=_("Employee"),
        help_text=_("The employee requesting the absence"),
    )

    notes = models.TextField(
        null=True, blank=True, verbose_name=_("Extra Notes"), help_text=_("A note to the HR administrator")
    )
    reason = models.TextField(
        null=True, blank=True, verbose_name=_("Reason"), help_text=_("The HR's response to this absence request")
    )
    created = models.DateTimeField(
        auto_now_add=True, verbose_name=_("Created"), help_text=_("The request creation time")
    )

    attachment = models.FileField(
        null=True,
        blank=True,
        max_length=256,
        verbose_name=_("Attachment"),
        upload_to="human_resources/absence_request/attachments",
        help_text=_("Upload a file to document this absence request (e.g. medical certificate)"),
    )

    crossborder_country = models.ForeignKey(
        to="geography.Geography",
        null=True,
        blank=True,
        related_name="absence_requests",
        on_delete=models.PROTECT,
        verbose_name=_("Crossborder Country"),
        help_text=_("The country where this absence request will be held."),
        limit_choices_to={"level": 1},
    )

    @transition(
        field=status,
        source=[Status.DRAFT],
        target=Status.PENDING,
        permission=lambda instance, user: user.profile == instance.employee.profile
        or user.has_perm("wbhuman_resources.administrate_absencerequest"),
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                color=ButtonDefaultColor.WARNING,
                identifiers=("wbhuman_resources:absencerequest",),
                icon=WBIcon.SEND.icon,
                key="submit",
                label=_("Submit"),
                action_label=_("Submitting"),
                description_fields=_("<p>Are you sure you want to submit this request?</p>"),
                instance_display=create_simple_display([["notes"]]),
            )
        },
    )
    def submit(self, **kwargs):
        pass

    def post_submit(self, **kwargs):
        msg = gettext(
            "<p>{employee} has submitted a {type} request for {hours} hours from {lower} to {upper}.</p>"
        ).format(
            employee=str(self.employee),
            type=str(self.type),
            hours=self.total_hours,
            lower=self.period.lower.strftime("%Y-%m-%d %H:%M:%S"),
            upper=self.period.upper.strftime("%Y-%m-%d %H:%M:%S"),
        )
        if self.crossborder_country:
            msg += gettext("</br><p><b>Country:</b></p><i>{0}</i>").format(str(self.crossborder_country))
        if self.notes and self.notes != "<p></p>" and self.notes != "null":
            msg += gettext("</br><p><b>Employee's Note:</b></p><i>{notes}</i>").format(notes=self.notes)
        title = gettext("New {type} Request").format(type=str(self.type))
        self.notify(title, msg, to_requester=False, to_manager=True)

    def can_submit(self):
        errors = dict()
        try:
            self.type.validate_country(self.crossborder_country)
        except ValueError as e:
            errors["crossborder_country"] = e.args[0]
        return errors

    @transition(
        field=status,
        source=[Status.PENDING],
        target=Status.APPROVED,
        on_error="failed",
        permission=lambda instance, user: can_validate_or_deny_request(instance, user),
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbhuman_resources:absencerequest",),
                icon=WBIcon.APPROVE.icon,
                color=ButtonDefaultColor.SUCCESS,
                key="approve",
                label=_("Approve"),
                action_label=_("Approval"),
                description_fields=_("<p>Are you sure you want to approve this request?</p>"),
            )
        },
    )
    def approve(self, **kwargs):
        pass

    def post_approve(self, **kwargs):
        msg = gettext("<p>Your {type} request from {start_date} to {end_date} has been approved.</p>").format(
            type=str(self.type),
            start_date=self.period.lower.strftime("%d.%m.%Y"),
            end_date=self.period.upper.strftime("%d.%m.%Y"),
        )
        title = gettext("Absence request approved")
        self.notify(title, msg, to_requester=True)

    @transition(
        field=status,
        source=[Status.PENDING],
        target=Status.DENIED,
        permission=lambda instance, user: can_validate_or_deny_request(instance, user),
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbhuman_resources:absencerequest",),
                icon=WBIcon.DENY.icon,
                color=ButtonDefaultColor.ERROR,
                key="deny",
                label=_("Deny"),
                action_label=_("Denial"),
                description_fields=_("<p>Are you sure you want to deny this request?</p>"),
                instance_display=create_simple_display([["reason"]]),
            )
        },
    )
    def deny(self, **kwargs):
        pass

    def post_deny(self, **kwargs):
        if self.type.is_vacation:
            msg = gettext("<p>Your absence request from {start_date} to {end_date} has been denied.</p>").format(
                start_date=self.period.lower.strftime("%d.%m.%Y"), end_date=self.period.upper.strftime("%d.%m.%Y")
            )
            if self.reason and self.reason not in ["<p></p>", ""]:
                msg += gettext("</br><p><b>HR's reason:</b></p><i>{reason}</i>").format(reason=self.reason)
            title = gettext("Absence request denied")
            self.notify(title, msg, to_requester=True)

    @transition(
        field=status,
        source=[Status.PENDING],
        target=Status.DRAFT,
        permission=lambda instance, user: user.profile == instance.employee.profile
        or user.has_perm("wbhuman_resources.administrate_absencerequest"),
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbhuman_resources:absencerequest",),
                color=ButtonDefaultColor.WARNING,
                icon=WBIcon.EDIT.icon,
                key="backtodraft",
                label=_("Back to Draft"),
                action_label=_("Back to Draft"),
                description_fields=_("<p>Are you sure you want to put this request back to draft?</p>"),
            )
        },
    )
    def backtodraft(self, **kwargs):
        pass

    @transition(
        field=status,
        source=[Status.APPROVED],
        target=Status.CANCELLED,
        on_error="failed",
        permission=can_cancel_request,
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbhuman_resources:absencerequest",),
                icon=WBIcon.REJECT.icon,
                color=ButtonDefaultColor.ERROR,
                key="cancel",
                label=pgettext_lazy("Transition button", "Cancel"),
                action_label=_("Cancellation"),
                description_fields=_("<p>Are you sure you want to cancel this request?</p>"),
            )
        },
    )
    def cancel(self, **kwargs):
        pass

    def post_cancel(self, **kwargs):
        msg = gettext("{employee} has cancelled a {type} request from {start_date} to {end_date}.").format(
            employee=str(self.employee),
            type=str(self.type),
            start_date=self.period.lower.strftime("%d.%m.%Y"),
            end_date=self.period.upper.strftime("%d.%m.%Y"),
        )
        self.periods.all().delete()
        title = gettext("Cancelled {type} Request").format(type=str(self.type))
        self.notify(title, msg, to_manager=True)

    objects = AbsenceRequestManager()

    class Meta:
        verbose_name = _("Absence Request")
        verbose_name_plural = _("Absence Requests")
        permissions = [("administrate_absencerequest", "Can Administrate Absence Requests")]

    @property
    @admin.display(description="Total hours")
    def total_hours(self) -> float:
        return getattr(self, "_total_hours", self.periods.aggregate(s=models.Sum("_total_hours"))["s"] or 0.0)

    @property
    @admin.display(description="Total Vacation hours")
    def total_vacation_hours(self) -> float:
        return getattr(
            self,
            "_total_vacation_hours",
            self.total_hours if self.type.is_vacation and self.status == AbsenceRequest.Status.APPROVED else 0.0,
        )

    @property
    @admin.display(description="Total hours (in days)")
    def total_hours_in_days(self) -> float:
        return getattr(self, "_total_hours_in_days", self.total_hours / self.employee.calendar.get_daily_hours())

    @property
    @admin.display(description="Total Vacation hours (in days)")
    def total_vacation_hours_in_days(self) -> float:
        return getattr(
            self, "_total_vacation_hours_in_days", self.total_vacation_hours / self.employee.calendar.get_daily_hours()
        )

    @property
    def next_extensible_period(self) -> TimestamptzRange | None:
        upper_date = (self.period.upper + BDay(1)).to_pydatetime()
        while self.employee.calendar.days_off.filter(date=upper_date.date(), count_as_holiday=True).exists():
            upper_date += timedelta(days=1)
        if not AbsenceRequestPeriods.objects.filter(employee=self.employee, date=upper_date).exists():
            return TimestamptzRange(lower=self.period.lower, upper=upper_date)

    def get_color(self) -> str:
        return self.type.color

    def get_icon(self) -> str:
        return self.type.icon

    def __str__(self) -> str:
        return f"{str(self.employee)} [{self.period.lower:%Y-%m-%d %H:%M}-{self.period.upper:%Y-%m-%d %H:%M}] ({self.Status[self.status].label})"

    def clean(self):
        if not self.period or not self.period.lower or not self.period.upper:
            raise ValidationError("Period needs to be set with nonnull bound")
        super().clean()

    def save(self, *args, **kwargs):
        self.title = f"{str(self.type)}: {str(self.employee)}"
        self.full_clean()
        if self.id and self.periods.count() > 1:
            self.all_day = True
        self.is_cancelled = self.status == self.Status.CANCELLED
        self.period = self.employee.calendar.normalize_period(
            self.period
        )  # We normalize the period to be sure it respect the default calendar working periods
        super().save(*args, **kwargs)

    def delete(self, **kwargs):
        super().delete(no_deletion=False)

    @property
    def periods_timespan(self) -> TimestamptzRange:
        """
        Property to returns the timespan datetime range originating from the attached periods

        Returns:
            A datetime range. None if no period are already attached to this request
        """
        periods = self.periods.order_by("timespan__startswith")
        if periods.exists():
            return TimestamptzRange(lower=periods.first().timespan.lower, upper=periods.last().timespan.upper)

    def notify(self, title: str, msg: str, to_requester: bool = True, to_manager: bool = False):
        """
        Get a message and a title and create the proper Notification object for the user administrating the HR module

        Args:
            title (str): The Notification title
            msg (str): The Notification message
        """
        users = []
        if to_requester and (requester_account := getattr(self.employee.profile, "user_account", None)):
            users.append(requester_account)
        if to_manager:
            for manager in self.employee.get_managers():
                if manager_account := getattr(manager, "user_account", None):
                    users.append(manager_account)
            for extra_group in self.type.extra_notify_groups.all():
                for extra_notify_user in extra_group.user_set.all():
                    users.append(extra_notify_user)
        for user in users:
            if user.is_active:
                send_notification(
                    code="wbhuman_resources.absencerequest.notify",
                    title=title,
                    body=msg,
                    user=user,
                    reverse_name="wbhuman_resources:absencerequest-detail",
                    reverse_args=[self.id],
                )

    def is_deletable_for_user(self, user: "User") -> bool:
        """
        Check if the given user has the permission to delete a request

        Args:
            user: Checked user

        Returns:
            True if the request can be deleted by the user
        """
        return self.is_deletable and (
            (self.status == AbsenceRequest.Status.PENDING and self.period.lower > timezone.now())
            or self.status == AbsenceRequest.Status.DRAFT
            or user.has_perm("wbhuman_resources.administrate_absencerequest")
        )

    @classmethod
    def get_endpoint_basename(cls) -> str:
        return "wbhuman_resources:absencerequest"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"

    @classmethod
    def get_representation_label_key(cls) -> str:
        return "{{employee}}"


class AbsenceRequestPeriodDefaultManager(models.Manager):
    def get_queryset(self) -> "models.QuerySet[AbsenceRequestPeriods]":
        """
        Default manager
        """
        return super().get_queryset().annotate(_total_hours=models.F("default_period__total_hours"))


class AbsenceRequestPeriodVacationManager(models.Manager):
    def get_queryset(self) -> "models.QuerySet[AbsenceRequestPeriods]":
        """
        Default manager
        """
        return (
            super()
            .get_queryset()
            .filter(request__type__is_vacation=True, request__status=AbsenceRequest.Status.APPROVED)
            .annotate(_total_hours=models.F("default_period__total_hours"))
        )


class AbsenceRequestPeriods(models.Model):
    request = models.ForeignKey(
        "wbhuman_resources.AbsenceRequest",
        related_name="periods",
        on_delete=models.CASCADE,
        verbose_name=_("Request"),
        help_text=_("The associated request"),
    )
    employee = models.ForeignKey(
        "wbhuman_resources.EmployeeHumanResource",
        related_name="periods",
        on_delete=models.CASCADE,
        verbose_name=_("Employee"),
        help_text=_("The Requester"),
    )  # Not supposed to be set dynamically. Use as a replicate to ensure database constraint
    default_period = models.ForeignKey(
        "wbhuman_resources.DefaultDailyPeriod",
        related_name="periods",
        on_delete=models.PROTECT,
        verbose_name=_("Period"),
        help_text=_("The associated period"),
    )
    date = models.DateField()

    timespan = DateTimeRangeField(
        verbose_name=_("Timespan"),
    )  # Expected to be read only

    balance = models.ForeignKey(
        "wbhuman_resources.EmployeeYearBalance",
        related_name="periods",
        on_delete=models.PROTECT,
        blank=True,
        null=True,
        verbose_name=_("Balance"),
        help_text=_("For which balance this absence will count towards"),
    )
    consecutive_hours_count = models.IntegerField(
        default=0,
        verbose_name=_("Consecutive Absence Hours count"),
        help_text=_("The number of consecutive hours this absence request period spans"),
    )

    objects = AbsenceRequestPeriodDefaultManager()
    vacation_objects = AbsenceRequestPeriodVacationManager()

    class Meta:
        verbose_name = _("Absence Request Period")
        verbose_name_plural = _("Absence Request Periods")
        unique_together = ("employee", "default_period", "date")
        indexes = [
            models.Index(fields=["employee", "default_period", "date"]),
        ]
        constraints = [
            ExclusionConstraint(
                name="exclude_overlapping_periods",
                expressions=[
                    ("timespan", RangeOperators.OVERLAPS),
                    ("employee", RangeOperators.EQUAL),
                ],
            ),
        ]

        notification_types = [
            create_notification_type(
                code="wbhuman_resources.absencerequest.notify",
                title="Absence Notification",
                help_text="Sends a notification when you can approve an absence request",
            )
        ]

    def __str__(self) -> str:
        return f"{self.request.employee.computed_str}: {self.timespan.lower:%Y-%m-%d %H:%M:%S}-{self.timespan.upper:%Y-%m-%d %H:%M:%S} ({self.total_hours})"

    @cached_property
    def total_hours(self) -> float:
        """
        A property holding the number of hours this period has. Get it from the attached default period
        """
        return getattr(self, "_total_hours", self.default_period.total_hours)

    @property
    def previous_period(self):
        """
        Return the previous approved period if it exists from this period's employee
        """
        return (
            AbsenceRequestPeriods.objects.exclude(
                request__status=AbsenceRequest.Status.CANCELLED,
            )
            .filter(
                employee=self.employee,
                timespan__startswith__lt=self.timespan.lower,
            )
            .order_by("-timespan__startswith")
            .first()
        )

    @classmethod
    def get_periods_as_df(cls, start: date, end: date, **extra_filter_kwargs) -> pd.DataFrame:
        """
        Utility function that reshape the periods for a given calendar and a list of employees among a certain date range

        Args:
            start: Start filter
            end: End filter
            extra_filter_kwargs: keyword argument that can be passed down to filter out employees

        Returns:
            A dataframe whose columns are [employee, period, date, type, status]
        """
        periods = cls.objects.filter(
            models.Q(date__gte=start),
            models.Q(date__lte=end),
            models.Q(_total_hours__gt=0)
            & models.Q(
                request__status__in=[
                    AbsenceRequest.Status.APPROVED.name,
                    AbsenceRequest.Status.PENDING.name,
                    AbsenceRequest.Status.DRAFT.name,
                ]
            ),
        ).filter(**extra_filter_kwargs)
        df_periods = pd.DataFrame(
            periods.values(
                "employee",
                "request__type__title",
                "request__status",
                "default_period",
                "date",
            ),
            columns=["employee", "request__type__title", "request__status", "default_period", "date"],
        ).rename(
            columns={
                "employee": "employee",
                "request__type__title": "type",
                "request__status": "status",
                "default_period": "period",
                "date": "date",
            }
        )
        return df_periods

    def assign_balance(self, check_date_availability: bool = True):
        """
        Utility function that returns what balance this request will be counted for

        Args:
            check_date_availability: If False, we bypass wether this request can be used toward the previous year balance from the prereference and only check if the previous year balance can accomotate the period's total hours
        """

        if self.request.type.is_vacation:
            previous_balances_with_positive_allowance = self.employee.balances.filter(
                _total_vacation_hourly_balance__gte=self.total_hours, year__lt=self.date.year
            )
            year = (
                previous_balances_with_positive_allowance.earliest("year").year
                if previous_balances_with_positive_allowance.exists()
                else self.date.year
            )

            # we loop to try to find, in order, the first balance to accomodate our request since the request's year to the max year
            while year <= self.date.year + 1 and self.balance is None:
                if possible_balance := self.employee.balances.filter(year=year).first():
                    if (
                        possible_balance.balance > 0
                        and possible_balance.total_vacation_hourly_balance >= self.total_hours
                        and get_previous_year_balance_expiration_date(possible_balance.year) > self.date
                    ):
                        self.balance = possible_balance
                        self.save()
                year += 1
            if self.balance is None:
                # we make sure that the max year is always the last seeded balance year + 1
                seeded_balances = self.employee.balances.filter(_given_balance__gt=0)
                if seeded_balances.exists():
                    self.balance = self.employee.get_or_create_balance(seeded_balances.latest("year").year + 1)[0]

    def get_consecutive_hours_count(self) -> float:
        """
        This subroutines aims to find the latest valid request days and get its consecutive days balance in order to
        compute this request day consecutive days balance.
        """
        consecutive_hours = self.total_hours
        if (
            (previous_period := self.previous_period)
            and previous_period.request.type.is_timeoff
            and not list(
                self.employee.extract_workable_periods(previous_period.timespan.upper, self.timespan.lower, count=1)
            )
        ):
            consecutive_hours += previous_period.consecutive_hours_count
        return consecutive_hours

    def save(self, *args, **kwargs):
        self.timespan = TimestamptzRange(
            lower=self.default_period.get_lower_datetime(self.date),
            upper=self.default_period.get_upper_datetime(self.date),
        )
        self.employee = self.request.employee
        self.consecutive_hours_count = self.get_consecutive_hours_count()
        super().save(*args, **kwargs)

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"

    @classmethod
    def get_representation_label_key(cls) -> str:
        return "{{request}}{{timespan}}{{_total_hours}}"


@receiver(post_save, sender=AbsenceRequest)
def post_save_absence_request(sender, instance, created, **kwargs):
    """
    Post save signal
    * Auto approve request if there are not absence request.
    * Compute and Create one AbsenceRequestPeriods if the request holds within the same year or two if the request spans multiple years.
    AbsenceRequestPeriods stores the total number of working days for a certain period of time.
    """

    if instance.type.auto_approve and instance.status == AbsenceRequest.Status.DRAFT:
        instance.submit()
        AbsenceRequest.objects.filter(id=instance.id).update(status=AbsenceRequest.Status.APPROVED)

    if employee := instance.employee:
        instance.entities.set([employee.profile])

    # We make sure that the periods are deleted on cancellation
    if instance.status == AbsenceRequest.Status.CANCELLED:
        instance.periods.all().delete()

    # Create for every workable period a Absence request period object for this request
    existing_periods = instance.periods
    for period_date, default_period in instance.employee.extract_workable_periods(
        instance.period.lower, instance.period.upper
    ):
        period, created = AbsenceRequestPeriods.objects.get_or_create(
            default_period=default_period, employee=instance.employee, date=period_date, defaults={"request": instance}
        )
        existing_periods = existing_periods.exclude(id=period.id)
    existing_periods.all().delete()
    update_kwargs = {"all_day": (instance.periods.count() > 1)}
    if (new_period := instance.periods_timespan) and new_period != instance.period:
        update_kwargs["period"] = new_period
    AbsenceRequest.objects.filter(id=instance.id).update(**update_kwargs)

    if instance.status == AbsenceRequest.Status.APPROVED and instance.type.is_vacation:
        for period in instance.periods.filter(balance__isnull=True).order_by("timespan__startswith"):
            period.assign_balance()


# TODO find an optimal way to recompute consecutive hours count on posterieurs periods save
# @shared_task
# def save_future_employee_periods_as_task(request_id: int, from_datetime: datetime):
#     request = AbsenceRequest.objects.get(id=request_id)
#     if (new_period := request.timespan) and new_period != request.period:
#         AbsenceRequest.objects.filter(id=request.id).update(period=new_period)
#     for period in AbsenceRequestPeriods.objects.filter(
#         employee=request.employee, timespan__startswith__gt=from_datetime
#     ).order_by("timespan__startswith"):
#         # If the consecutive hours is equals to a normal working day, we assume this is the beginning of a new continious serie of periods
#         if period.consecutive_hours_count == period.default_period.calendar.get_daily_hours():
#             break
#         period.save()


# @receiver(post_delete, sender=AbsenceRequestPeriods)
# @receiver(post_save, sender=AbsenceRequestPeriods)
# def receiver_absence_request_periods(sender, instance, **kwargs):
#     if kwargs.get("created", False) or not hasattr(kwargs, "created"):
#         save_future_employee_periods_as_task.delay(instance.request.id, instance.timespan.upper)


@receiver(post_save, sender=DayOff)
def post_save_dayoff(sender, instance, created, **kwargs):
    """
    Post save signal, Ensure that when a day off is created, all absence request periods are recomputed
    """
    if created:
        for period in AbsenceRequestPeriods.objects.filter(date=instance.date):
            period.request.save()


@receiver(post_save, sender=AbsenceRequestType)
def post_save_absence_request_type(sender, instance: AbsenceRequestType, created: bool, raw: bool, **kwargs):
    """
    Post save signal for absence request type
    """
    # Ensure that when a absence request where country validation is necessary, all countries are added to the allowed list by default
    if created and instance.is_country_necessary:
        instance.crossborder_countries.set(Geography.countries.all())

    if not raw:
        # We need to trigger all requests' save methods to update their color and icon
        for request in instance.request.all():
            request.save()


@receiver(post_save, sender=Geography)
def post_country_creation(sender, instance, created, **kwargs):
    """
    Post save signal, Ensure that when a country is created, it is added by default to all absence request where country validation is necessary
    """
    if created and instance.level == 1:
        for absence_request_type in AbsenceRequestType.objects.filter(is_country_necessary=True):
            absence_request_type.crossborder_countries.add(instance)



# Generated by Django 4.1.9 on 2023-06-08 09:58

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("auth", "0012_alter_user_first_name_max_length"),
        ("geography", "0001_initial"),
        ("wbhuman_resources", "0016_alter_employeehumanresource_options"),
    ]

    operations = [
        migrations.AddField(
            model_name="absencerequest",
            name="crossborder_country",
            field=models.ForeignKey(
                blank=True,
                help_text="The country where this absence request will be held.",
                limit_choices_to={"level": 1},
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="absence_requests",
                to="geography.geography",
                verbose_name="Crossborder Country",
            ),
        ),
        migrations.AddField(
            model_name="absencerequesttype",
            name="crossborder_countries",
            field=models.ManyToManyField(
                blank=True,
                help_text="List of countries where crossborder activity is allowed",
                limit_choices_to={"level": 1},
                to="geography.geography",
                verbose_name="Countries",
            ),
        ),
        migrations.AddField(
            model_name="absencerequesttype",
            name="extra_notify_groups",
            field=models.ManyToManyField(
                blank=True,
                related_name="notified_absence_request_types",
                to="auth.group",
                verbose_name="Extra Notify Groups",
            ),
        ),
        migrations.AddField(
            model_name="absencerequesttype",
            name="is_country_necessary",
            field=models.BooleanField(default=False, verbose_name="Is country necessary"),
        ),
    ]



from rest_framework.reverse import reverse
from wbcore.metadata.configs.endpoints import EndpointViewConfig


class AbsenceRequestEmployeeHumanResourceEndpointConfig(EndpointViewConfig):
    def get_create_endpoint(self, **kwargs):
        return reverse("wbhuman_resources:absencerequest-list", args=[], request=self.request)

    def get_endpoint(self, **kwargs):
        return reverse(
            "wbhuman_resources:employee-absencerequest-list",
            args=[self.view.kwargs["employee_id"]],
            request=self.request,
        )


class YearBalanceEmployeeHumanResourceEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return reverse(
            "wbhuman_resources:employee-employeeyearbalance-list",
            args=[self.view.kwargs["employee_id"]],
            request=self.request,
        )

    def get_create_endpoint(self, **kwargs):
        return None

    def get_delete_endpoint(self, **kwargs):
        return None


class EmployeeBalanceEndpointConfig(EndpointViewConfig):
    def get_instance_endpoint(self, **kwargs):
        return reverse("wbhuman_resources:employeebalance-list", args=[], request=self.request)

    def get_endpoint(self, **kwargs):
        return None


class EmployeeEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return None


class WeeklyOffPeriodEmployeeHumanResourceEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return reverse(
            "wbhuman_resources:employee-weeklyoffperiod-list",
            args=[self.view.kwargs["employee_id"]],
            request=self.request,
        )



from django.utils.translation import gettext_lazy as _
from wbcore import filters as wb_filters


def add_position_filter(sender, request=None, *args, **kwargs):
    from wbcore.contrib.directory.models import Entry

    from wbhuman_resources.models import Position

    def filter_position(queryset, name, value):
        if value:
            ids = value.get_employees().values_list("profile")
            entries = Entry.objects.filter(id__in=ids)
            return queryset.filter(participants__in=entries)
        return queryset

    return {
        "position": wb_filters.ModelChoiceFilter(
            label=_("Participant's position"),
            field_name="position",
            queryset=Position.objects.all(),
            endpoint=Position.get_representation_endpoint(),
            value_key=Position.get_representation_value_key(),
            label_key=Position.get_representation_label_key(),
            method=filter_position,
        )
    }
